#include <stdarg.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <stdio.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/select.h>
#include <sys/wait.h>
#include <sprog.h>

void sprog_process(const struct sprog_family *fam, void *arg, const struct sprog_data *d, int nstdin, int nstdout);

void sprog_error(const char *text, ...) {
  va_list l;
  va_start(l, text);
  vfprintf(stderr, text, l);
  va_end(l);
}

void sprog_sleep(int msec) {
  usleep(msec * 1000);
}

void sprog_communicate(const struct sprog_family *fam, const struct sprog_data *d, const char *port, int baud) {
  void *arg;
  struct serial_device dev;
  
  serial_open(&dev, port, baud);
  arg = fam->setup(&dev);
  sprog_process(fam, arg, d, dev.fd, dev.fd);
}

int sprog_waitdata(int timeout) {
  fd_set read_set;
  struct timeval tval;
  FD_ZERO(&read_set);
  FD_SET(1, &read_set);
  tval.tv_sec = timeout/1000;
  tval.tv_usec = (timeout%1000)*1000;
  if(select(1+1, &read_set, NULL, NULL, &tval)!=1)
    return 0;
  return 1;
}

void sprog_process(const struct sprog_family *fam, void *arg, const struct sprog_data *d, int nstdin, int nstdout) {
  int oldstdin;
  int oldstdout;
  oldstdin = dup(0);
  oldstdout = dup(1);
  dup2(nstdin, 0);
  dup2(nstdout, 1);
  fam->init(arg);
  if(d)
    fam->exec(arg, d);
  dup2(oldstdin, 0);
  dup2(oldstdout, 1);
}

void sprog_load(struct sprog_data *d, const char *path) {
  char buf[4096];
  int fd;
  int n;
  
  fd = open(path, O_RDONLY);
  if(fd<0) {
    sprog_error("Unable to open '%s': %s\n", path, strerror(errno));
    return;
  }
  
  d->size = 0; 

  while((n = read(fd, buf, sizeof(buf)))>0)
    sprog_append_data(d, buf, n);
    
  if(n<0) {
    sprog_error("Error while reading from '%s': %s\n", path, strerror(errno));
    free(d->data);
    d->size = 0;
    d->data = NULL;
    return;
  }
}

void sprog_append_data(struct sprog_data *d, const char *data, int len) {
  int i;
  int alloc;
  alloc = d->alloc;
  
  if(!d->data)
    alloc = 4096;
  
  while((d->size+len+128)<alloc)
    alloc *= 2;
  
  if(d->data)
    d->data = realloc(d->data, alloc);
  else
    d->data = malloc(alloc);
  
  for(i=0; i<len; i++)
    d->data[i+d->size] = data[i];
  
  d->alloc = alloc;
  d->size = len;
}

void sprog_progress(int progress) {
  char buf[80];
  int i;
  int w;
  w = (progress*7)/10;
  for(i=0; i<70; i++) {
    if(i<=w)
      buf[i] = ':';
    else
      buf[i] = ' ';
  }
  buf[i] = 0;
  
  sprog_error("\r[%s] %3d%%", buf, progress);
}
 
